#include <libps.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>

#include "connect4.h"
#include "pnsearch.h"
#include "proto.h"

#include "pad.h"

extern struct board *the_board;

typedef void (*callbackFN) (void);

callbackFN draw_func;
callbackFN pad_func;

#define HEAP_START (0x80180000)
#define HEAP_SIZE (0x7F000)

#define TIM_ADDR (0x80180000)
#define VH_ADDR (0x800AEF00)
#define VB_ADDR (0x80193000)

#define OCTAVE_FACTOR 24

#define SCREEN_WIDTH 384
#define SCREEN_HEIGHT 256
#define TITLE 0
#define STARTING 1
#define RUNNING 2
#define ENDING 3
#define MIN_X 60
#define MIN_Y 40
#define COUNTER_SIZE 32
#define EMPTY 0
#define BLUE 1
#define RED 2
#define C4_DRAW 3
#define RED_OFFSET 0
#define BLUE_OFFSET (COUNTER_SIZE)
#define BOARD_OFFSET (2*COUNTER_SIZE)
#define MAX_X 7
#define MAX_Y 6
#define HORIZONTAL 1
#define VERTICAL 2
#define DIAGONAL1 4
#define DIAGONAL2 8
#define DROP_SPEED 1
#define DROP_ACCEL 1.2
#define FLASH_TIME 15
#define FLASH_OFF_TIME 10
#define MOVE_PAUSE 10

/* variables for veleng */
int move_count;
char move_list[50];

char red_ai= '0';
char blue_ai= '0';

int computer_move;
int computer_thinking;

int frames_until_next_action;

int choosing_ai;

u_long last_pad= 0;

int left_abort=0;
int right_abort=0;

int flash_until= 0;
int win_line;
int move_x;
int move_y;
int not_quitting= 1;
int game_status= TITLE;
int current_frame= 0;
int win_font;
int in1_font;
int in2_font;
int in3_font;
int in4_font;
int output_buffer_index;
GsOT world_ordering_table[2];
GsOT_TAG ordering_table[2][1<<1];
PACKET gpu_work_area[2][20000];
GsSPRITE counter_sprite;
GsIMAGE tim;
int winner;
int current;
int grid_count;
int curr_x;
int curr_y;
int dropping_counter;
int target_y;
int drop_speed;
float flt_drop_speed;
short vab_voice;

int grid[MAX_X][MAX_Y];

char *get_ai_string (char ai_ch)
{
  switch (ai_ch) {
    case '0':
      return "human";
    case 'a':
      return "weak";
    case 'b':
      return "normal";
    default:
      return "strong";
  }
}


void *c4_malloc (int size)
{
  void *result;

  result= malloc (size);

  if (!result) printf ("malloc call failed: size is %d\n",
		size);

  return result;
}

void dummy_fn (void)  {
}

void load_tim (u_long *addr, GsIMAGE *tim)
{
  RECT rect;

  addr ++;
  GsGetTimInfo (addr, tim);
  DrawSync (0);
  rect.x= tim->px;
  rect.y= tim->py;
  rect.w= tim->pw;
  rect.h= tim->ph;
  LoadImage (&rect, tim->pixel);
  DrawSync (0);

  if ((tim->pmode >> 3) & 0x01) {
    rect.x= tim->cx;
    rect.y= tim->cy;
    rect.w= tim->cw;
    rect.h= tim->ch;
    LoadImage (&rect, tim->clut);
  }
  DrawSync (0);
}

/* load tim data into sprites */
void load_sprite (GsSPRITE *spr, GsIMAGE *tim)
{
  spr->attribute= (1<<24)|(1<<30)|(1<<29);
  spr->w= tim->pw*2;
  spr->h= tim->ph;
  spr->tpage= GetTPage (1,0,tim->px,tim->py);
  spr->u= 0;
  spr->v= 0;
  spr->cx= tim->cx;
  spr->cy= tim->cy;
  spr->r= 0x80;
  spr->g= 0x80;
  spr->b= 0x80;
  spr->mx= tim->pw;
  spr->my= tim->ph;
  spr->scalex= ONE;
  spr->scaley= ONE;
  spr->rotate= 0;
}


void reset_grid (void) {
  int i, j;

  for (i=0; i<MAX_X; i++)
    for (j=0; j<MAX_Y; j++)
      grid[i][j]= EMPTY;
}

int check_move (int col, int x, int y) {
  int curr;
  int count;
  int i;
  int rv= 0;

  grid[x][y]= col;

  count= 0;
  for (curr= x; curr >= 0; curr --) {
    if (grid[curr][y]==col) count ++;
    else break;
    if (curr==x-3) break;
  }

  count --;
  for (curr= x; curr < MAX_X; curr ++) {
    if (grid[curr][y]==col) count ++;
    else break;
    if (curr==x+3) break;
  }

  if (count >= 4) rv= HORIZONTAL;

  count= 0;
  for (curr= y; curr >= 0; curr --) {
    if (grid[x][curr]==col) count ++;
    else break;
    if (curr==y-3) break;
  }

  count --;
  for (curr= y; curr < MAX_Y; curr ++) {
    if (grid[x][curr]==col) count ++;
    else break;
    if (curr==y+3) break;
  }

  if (count >= 4) rv= rv|VERTICAL;

  count= 0;
  i= y;
  for (curr= x; curr >= 0; curr --) {
    if (grid[curr][i]==col) count ++;
    else break;
    if (curr==x-3 || i==0) break;
    i --;
  }

  count --;
  i= y;
  for (curr= x; curr < MAX_X; curr ++) {
    if (grid[curr][i]==col) count ++;
    else break;
    if (curr==x+3 || i==MAX_Y-1) break;
    i ++;
  }

  if (count >=4) rv= rv|DIAGONAL1;


  count= 0;
  i= y;
  for (curr= x; curr >= 0; curr --) {
    if (grid[curr][i]==col) count ++;
    else break;
    if (curr==x-3 || i==MAX_Y-1) break;
    i ++;
  }

  count --;
  i= y;
  for (curr= x; curr < MAX_X; curr ++) {
    if (grid[curr][i]==col) count ++;
    else break;
    if (curr==x+3 || i==0) break;
    i --;
  }

  if (count >=4) rv= rv|DIAGONAL2;

  return rv;
}

int find_y (int x) {
  int i= MAX_Y;
  int not_finished= 1;

  if (x<0 || x>=MAX_X) return MAX_Y;

  while (not_finished) {
    if (grid[x][i-1] == EMPTY) not_finished= 0;
    i--;

    if (i==0) {
      if (not_finished)
	i= MAX_Y;
      not_finished= 0;
    }
  }


  return i;
}

void display_grid (void) {
  int i, j;
  char buff[80];

  if (computer_thinking) {
    sprintf (buff, "Veleng is thinking...");
  }

  if (left_abort) {
    sprintf (buff, "Abort requested by Red");
  }

  if (right_abort) {
    sprintf (buff, "Abort requested by Blue");
  }

  if (left_abort&&right_abort) {
    sprintf (buff, "Game aborted");
  }

  if (choosing_ai) {
    if (choosing_ai==RED)
      sprintf (buff, "Red player: %s", get_ai_string (red_ai));

    if (choosing_ai==BLUE)
      sprintf (buff, "Blue player: %s", get_ai_string (blue_ai));
  }

  if (left_abort||right_abort||computer_thinking||choosing_ai) {
    FntPrint (win_font, buff);
    FntFlush (win_font);
  }

  counter_sprite.v= BOARD_OFFSET;

  for (j= 0; j< MAX_Y; j ++)
    for (i= 0; i< MAX_X; i ++) {
      counter_sprite.x= MIN_X+COUNTER_SIZE*i;
      counter_sprite.y= MIN_Y+COUNTER_SIZE*j;
      GsSortFastSprite (&counter_sprite, &world_ordering_table[output_buffer_index], 0);
    }


  for (j= MAX_Y-1; j>=0; j--) {
    for (i= 0; i<MAX_X; i++) {
      switch (grid[i][j]) {
	case RED:
	  counter_sprite.x= MIN_X+COUNTER_SIZE*i;
	  counter_sprite.y= MIN_Y+COUNTER_SIZE*j;
	  counter_sprite.v= RED_OFFSET;
	  GsSortFastSprite (&counter_sprite, &world_ordering_table[output_buffer_index], 0);
	  break;
	case BLUE:
	  counter_sprite.x= MIN_X+COUNTER_SIZE*i;
	  counter_sprite.y= MIN_Y+COUNTER_SIZE*j;
	  counter_sprite.v= BLUE_OFFSET;
	  GsSortFastSprite (&counter_sprite, &world_ordering_table[output_buffer_index], 0);
	  break;
	default:
      }
    }
  }

  if (current != EMPTY) {
    counter_sprite.x= curr_x;
    counter_sprite.y= curr_y;
    if (current==RED)
      counter_sprite.v= RED_OFFSET;
    else
      counter_sprite.v= BLUE_OFFSET;
    GsSortFastSprite (&counter_sprite, &world_ordering_table[output_buffer_index], 0);
  }
}

void show_win (int x, int y, int win_line)
{
  int col;
  int xl[30];
  int yl[30];
  int i;
  int ii;
  int curr;
  int i_count;

  i= 0;
  col= grid[x][y];

  if (win_line & HORIZONTAL) {
    for (curr= x; curr >= 0; curr --) {
      if (grid[curr][y]==col) {
	xl[i]= curr;
	yl[i]= y;
	i++;
      } else break;
      if (curr==x-3) break;
    }

    for (curr= x; curr < MAX_X; curr ++) {
      if (grid[curr][y]==col) {
	xl[i]= curr;
	yl[i]= y;
	i ++; 
      } else break;
      if (curr==x+3) break;
    }
  }
  if (win_line & VERTICAL) {
    for (curr= y; curr >= 0; curr --) {
      if (grid[x][curr]==col) {
	xl[i]= x;
	yl[i]= curr;
	i++;
      } else break;
      if (curr==y-3) break;
    }

    for (curr= y; curr < MAX_Y; curr ++) {
      if (grid[x][curr]==col) {
	xl[i]= x;
	yl[i]= curr;
	i ++; 
      } else break;
      if (curr==y+3) break;
    }
  }

  if (win_line & DIAGONAL1) {
    ii= y;
    for (curr= x; curr >= 0; curr --) {
      if (grid[curr][ii]==col) {
	xl[i]= curr;
	yl[i]= ii;
	i ++;
      } else break;
      if (curr==x-3 || ii==0) break;
      ii --;
    }

    ii= y;
    for (curr= x; curr < MAX_X; curr ++) {
      if (grid[curr][ii]==col) {
	xl[i]= curr;
	yl[i]= ii;
	i ++;
      } else break;
      if (curr==x+3 || ii==MAX_Y-1) break;
      ii ++;
    }
  }

  if (win_line & DIAGONAL2) {
    ii= y;
    for (curr= x; curr >= 0; curr --) {
      if (grid[curr][ii]==col) {
	xl[i]= curr;
	yl[i]= ii;
	i ++;
      } else break;
      if (curr==x-3 || ii==MAX_Y-1) break;
      ii ++;
    }

    ii= y;
    for (curr= x; curr < MAX_X; curr ++) {
      if (grid[curr][ii]==col) {
	xl[i]= curr;
	yl[i]= ii;
	i ++;
      } else break;
      if (curr==x+3 || ii==0) break;
      ii --;
    }
  }


  i --;

  i_count= i;

  if (current_frame>=flash_until) {
    while (i>=0) {
      grid[xl[i]][yl[i]]= EMPTY;
      i --;
    }
  }

  if (current_frame>=flash_until+FLASH_OFF_TIME)
    flash_until= current_frame + FLASH_TIME;

  display_grid ();

  i= i_count;
  while (i>=0) {
    grid[xl[i]][yl[i]]= col;
    i --;
  }
}

void title_draw (void)
{
  char buff[20];

  buff[0]= NULL;

  if (winner==RED) sprintf (buff, "Red wins");
  if (winner==BLUE) sprintf (buff, "Blue wins");
  if (winner==C4_DRAW) sprintf (buff, "Draw");

/* this font gets overwritten by display_grid if necessary */
  FntPrint (win_font, buff);
  FntFlush (win_font);

  if (flash_until<current_frame && (winner==RED || winner==BLUE)) {
    show_win (move_x, move_y, win_line);
  } else display_grid ();

  FntPrint (in1_font, "Press start to play");
  FntFlush (in1_font);

  FntPrint (in2_font, "Use left and right to change column");
  FntFlush (in2_font);

  FntPrint (in3_font, "Use cross to choose column");
  FntFlush (in3_font);

  FntPrint (in4_font, "Use triangle for AI options");
  FntFlush (in4_font);
}

void start_game (void)
{
  current_frame= 0;

  move_count= 0;
  move_list[0]= 'c';
  move_list[1]= '0';

  frames_until_next_action= 0;

  computer_thinking= 0;
  computer_move= 0;

  choosing_ai= 0;

  reset_grid ();
  grid_count= MAX_X*MAX_Y;
  dropping_counter= 0;
  winner= EMPTY;
  curr_x= MIN_X;
  move_x= 0;
  curr_y= MIN_Y - COUNTER_SIZE;
  left_abort= 0;
  right_abort= 0;
}

void choose_ai (u_long pad_status)
{
  if ((pad_status & (PADLleft|PADLright)) &&
	  (!(last_pad & (PADLleft|PADLright)))) {
    if (choosing_ai==BLUE) choosing_ai= RED;
    else choosing_ai= BLUE;
    SsUtKeyOn (vab_voice, 0, 0, 60-OCTAVE_FACTOR, 0, 127,127);
  }

  if ((pad_status & PADLdown) &&
	  !(last_pad & PADLdown)) {
    if (choosing_ai==BLUE) {
      blue_ai ++;
      if (blue_ai=='1') blue_ai= 'a';
      if (blue_ai=='d') blue_ai= '0';
    } else {
      red_ai ++;
      if (red_ai=='1') red_ai= 'a';
      if (red_ai=='d') red_ai= '0';
    }

    SsUtKeyOn (vab_voice, 1, 0, 60-OCTAVE_FACTOR, 0, 127,127);
  }

  if ((pad_status & PADLup) &&
	  !(last_pad & PADLup)) {
    if (choosing_ai==BLUE) {
      blue_ai --;
      if (blue_ai+1=='0') blue_ai= 'c';
      if (blue_ai+1=='a') blue_ai= '0';
    } else {
      red_ai --;
      if (red_ai+1=='0') red_ai= 'c';
      if (red_ai+1=='a') red_ai= '0';
    }

    SsUtKeyOn (vab_voice, 1, 0, 60-OCTAVE_FACTOR, 0, 127,127);
  }

/* exit AI choice */
  if ((pad_status & PADRup) && !(last_pad & PADRup))
    choosing_ai= 0;
}

void title_pad (void) {
  u_long pad_status;

  pad_status= PadRead ();

  if (choosing_ai)
    choose_ai (pad_status);

  if (pad_status & (PADLleft|PADLright|PADLup|PADLdown|PADRup)) {
    if (!choosing_ai)
      choosing_ai= BLUE;
  }

  if (pad_status & PADstart) {
    if (pad_status & PADselect)
      not_quitting= 0;
    else {
      current= RED;
      game_status= STARTING;
    }
  }

  if (pad_status & PAD2start) {
    if (pad_status & PAD2select) {
      not_quitting= 0;
    } else {
      current= BLUE;
      game_status= STARTING;
    }
  }

  last_pad= pad_status;
}

void move_left (void)
{
  SsUtKeyOn (vab_voice, 0, 0, 60-OCTAVE_FACTOR, 0, 127,127);
  move_x --;
  curr_x -= COUNTER_SIZE;
}

void move_right (void)
{
  SsUtKeyOn (vab_voice, 0, 0, 60-OCTAVE_FACTOR, 0, 127,127);
  move_x ++;
  curr_x += COUNTER_SIZE;
}

void drop_counter (void)
{
  dropping_counter= 1;
  drop_speed= DROP_SPEED;
  flt_drop_speed= DROP_SPEED;
  target_y= MIN_Y+COUNTER_SIZE*move_y;
}

void get_move (void)
{
  u_long pad_status;
  u_long this_left, this_right, this_drop;

/* human input - parse every frame so aborts are picked up*/
  if (current==RED) {
    this_left= PADLleft;
    this_right= PADLright;
    this_drop= PADRdown;
  } else {
    this_left= PAD2Lleft;
    this_right= PAD2Lright;
    this_drop= PAD2Rdown;
  }

  pad_status= PadRead ();

  if (pad_status & PADselect) {
    left_abort= 1;
    if (red_ai != '0' || blue_ai != '0')
      right_abort= 1;
  }

  if (pad_status & PAD2select) {
    right_abort= 1;
    if (red_ai != '0' || blue_ai != '0')
      left_abort= 1;
  }

  if (left_abort && right_abort)
    game_status= ENDING;

  if ((pad_status & PADstart) && (pad_status & PADselect))
    not_quitting= 0;
  if ((pad_status & PAD2start) && (pad_status & PAD2select))
    not_quitting= 0;

  if (choosing_ai) {
    choose_ai (pad_status);
  } else {
    if ((pad_status & PADRup) && !(last_pad & PADRup)) {
      if (!choosing_ai)
        choosing_ai= BLUE;
    }
  }

/* check if we should be pausing */
  if (frames_until_next_action<=current_frame && !choosing_ai) {
    if (!dropping_counter) {
      if ((current==RED && red_ai != '0') ||
	  (current==BLUE && blue_ai != '0')) {
/* computer plays */
	if (computer_move==0) {
/* pause 1 frame to print "thinking..." message */
	  computer_thinking= 1;
	  computer_move= -1;
	} else {
	  if (computer_move==-1) {
/* compute next move */
/* set correct ai level */
	    if (current==RED) move_list[0]= red_ai;
	    else move_list[0]= blue_ai;

/* get move from veleng */
	    computer_move= playgame (move_list, the_board);
	    computer_thinking= 0;
	  }

	  if (computer_move > 0) {
/* move to correct postion */
	    if (move_x > (computer_move-1)) {
	      move_left ();
	      frames_until_next_action= current_frame+MOVE_PAUSE;
	    } else if (move_x < (computer_move-1)) {
	      move_right ();
	      frames_until_next_action= current_frame+MOVE_PAUSE;
	    } else {
/* in the correct column */
	      move_y= find_y (move_x);
	      drop_counter ();
/* reset ai flag */
	      computer_move= 0;
	    }
	  }
	}
      } else if ((pad_status & this_left) && !(last_pad & this_left)) {
/* parse input */
        if (move_x>0) {
	  move_left ();
        }
      } else if ((pad_status & this_right) && !(last_pad & this_right)) {
        if (move_x<MAX_X-1) {
	  move_right ();
        }
      } else if (pad_status & this_drop) {
        move_y= find_y (move_x);
        if (move_y != MAX_Y) {
	  drop_counter ();
        }
      }
    }
  }

  if (dropping_counter) {
/* drop counter */
    curr_y += drop_speed;
    flt_drop_speed *= DROP_ACCEL;
    drop_speed= (int) flt_drop_speed;

/* check if hit bottom */
    if (curr_y>=target_y) {
/* update move list */
      move_count ++;
      move_list[move_count]= '1' + move_x;
      move_list[move_count+1]= '0';

      SsUtKeyOn (vab_voice, 1, 0, 60-OCTAVE_FACTOR, 0, 127,127);
      dropping_counter= 0;
      grid_count --;
      curr_y= MIN_Y - COUNTER_SIZE;
      if ((win_line= check_move (current, move_x, move_y))) {
	SsUtKeyOn (vab_voice, 2, 0, 60-OCTAVE_FACTOR, 0, 127, 127);
        winner= current;
      } else if (grid_count==0) {
	SsUtKeyOn (vab_voice, 2, 0, 60-OCTAVE_FACTOR, 0, 127, 127);
	winner= C4_DRAW;
      }

      if (winner!=EMPTY) {
	game_status= ENDING;
	current= EMPTY;
	flash_until= current_frame + FLASH_TIME;
      }

      if (current==RED) current= BLUE;
      else if (current==BLUE) current= RED;
    }
  }

  last_pad= pad_status;
}

int main (void)
{
  int count;

  PadInit ();

  SetVideoMode (MODE_PAL);
  ResetGraph (0);
  GsInitGraph (SCREEN_WIDTH, SCREEN_HEIGHT, GsOFSGPU, 0, 0);
  GsDefDispBuff (0, 0, 0, 256);
  GsDISPENV.screen.x= 0;
  GsDISPENV.screen.y= 18;
  GsDISPENV.screen.w= SCREEN_WIDTH;
  GsDISPENV.screen.h= SCREEN_HEIGHT;

  GsDRAWENV.r0= 0x00;
  GsDRAWENV.g0= 0x00;
  GsDRAWENV.b0= 0x00;
  GsDRAWENV.isbg= 1;

  for (count= 0; count<2; count ++) {
    world_ordering_table[count].length= 1;
    world_ordering_table[count].org= ordering_table[count];
    GsClearOt (0, 0, &world_ordering_table[count]);
  }

  vab_voice= SsVabTransfer ((unsigned char *)VH_ADDR, (unsigned char *)VB_ADDR, 1, 1);

  load_tim ((u_long *) TIM_ADDR, &tim);
  load_sprite (&counter_sprite, &tim);
  counter_sprite.h= COUNTER_SIZE;

  FntLoad (960, 256);
  win_font= FntOpen (100, 10, SCREEN_WIDTH, 8, 0, 40);

  in1_font= FntOpen (100, 50, 180, 8, 1, 21);
  in2_font= FntOpen (50, 90, 340, 8, 1, 40);
  in3_font= FntOpen (70, 110, 210, 8, 1, 40);
  in4_font= FntOpen (68, 130, 220, 8, 1, 40);

  draw_func= title_draw;
  pad_func= title_pad;

  InitHeap (HEAP_START, HEAP_SIZE);

  setup_veleng ();

  choosing_ai= 0;

  while (not_quitting) {
    current_frame ++;

    if (game_status==STARTING) {
      start_game ();

      draw_func= display_grid;
      pad_func= get_move;
      game_status= RUNNING;
    } else if (game_status==ENDING) {
      draw_func= title_draw;
      pad_func= title_pad;
      game_status= TITLE;
      choosing_ai= 0;
    }

    pad_func ();

    output_buffer_index= GsGetActiveBuff ();
    GsSetWorkBase ((PACKET *)gpu_work_area[output_buffer_index]);
    GsClearOt (0, 0, &world_ordering_table[output_buffer_index]);

    draw_func ();

    VSync (0);
    GsSwapDispBuff ();
    GsDrawOt (&world_ordering_table[output_buffer_index]);
  }

  SsUtAllKeyOff (0);
  SsVabClose (vab_voice);
  ResetGraph (1);

  close_veleng ();
  return 0;
}



